from ..context import get_context
from functools import partial
import asyncio
import copy

context_datastore_key = "datastreams"


def get_datastream(id, context=None):
    """
    Retrieves a :meth:`Datastream` instance for the given Datastream ID, creating
    one if nonexistant.

    This method is thread-safe.

    :param id: The ID of the Datastream instance to get, or to create.
    :param context: The optional context to use. If None, then this will use the
        context registered for the current thread.

    :returns: An instance of :class:`Datastream`
    """
    if context is None:
        context = get_context()
    datastreams_data, lock = context.get_interface_data(context_datastore_key)
    with lock:
        if id not in datastreams_data:
            datastreams_data[id] = Datastream()
        return datastreams_data[id]


class Datastream(object):
    """
    Provides an interface to share data, in the form of a dictionary, between
    different modules, with mechanisms for thread-safe operation, and for
    triggering asyncio events when certain conditions are met.
    """

    def __init__(self):
        self.event_triggers = list()
        self.data = dict()

    def push(self, data):
        """
        Saves data to the Datastream

        :param data: The dictionary of data to :meth:`dict.update()` the existing
         dictionary with.
        """
        self.data.update(data)
        self._check_events()

    def push_threadsafe(self, data, context=None):
        """
        Schedules :meth:`.push()` to be run in the current context's event loop.

        This method is thread-safe.

        :param data: The dictionary of data to give to :meth:`.push()`
        :param context: The (optional) context to use for scheduling. Otherwise
            the current thread's context will be used.
        """
        if context is None:
            context = get_context()
        context.get_event_loop().call_soon_threadsafe(partial(self.push, data))

    def get(self):
        """
        Retrieves data from the Datastream

        :returns: A copy of the contained data.
        """
        return self.data.copy()

    def _check_events(self):
        """Loops through all events, setting and clearing appropriately."""
        for event in self.event_triggers:
            if event["conditional"](self.data):
                event["event"].set()
            else:
                event["event"].clear()

    def set_event(self, conditional, event=None):
        """
        Adds a conditional to be evaluated at each :meth:`push()`. The event will be triggered
        when the conditional returns true, and cleared when the conditional returns false.

        :param conditional: A reference to a callable that accepts one argument,
            which is the current data dictionary held in the datastream.
        :param event: The optional event object to use. If none is provided, a new
            asyncio event will be created.

        :returns: The event that will be triggered when the conditional returns True.
        """
        if event is None:
            event = asyncio.Event()
        self.event_triggers.append({"event": event, "conditional": conditional})

        #Update events.
        self._check_events()
        return event

    def drop_event(self, event):
        """
        Remove event and it's attached conditional from being triggered.

        :param event: The event object to remove.

        :returns: True if successful.
        """
        for trigger in self.event_triggers[:]:
            if event is trigger["event"]:
                self.event_triggers.remove(event)
                return True
        return False


"""
This is an interface designed to make it easier to communicate
gamemode between modules.
"""

from ..context import get_context
import asyncio
import wpilib

context_data_key = "gamemode"

DISABLED = 0
TELEOPERATED = 1
AUTONOMOUS = 2


def set_gamemode(mode, context=None):
    """
    Sets the current gamemode and releases any coroutines waiting for it.

    :param mode: The gamemode to set.
    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """
    if context is None:
        context = get_context()
    data, lock = context.get_interface_data(context_data_key)
    with lock:
        if mode != data.get("mode", " "):
            data["last_change"] = wpilib.Timer.getFPGATimestamp()
        data["mode"] = mode
        context.thread_coroutine(_on_gamemode_set(mode, context))


@asyncio.coroutine
def _on_gamemode_set(mode, context):
    """
    Sets all saved events for the given gamemode
    """
    data, lock = context.get_interface_data(context_data_key)
    events = data.get("events-mode-" + str(mode), None)
    if events is not None:
        for event in events:
            event.set()

#Gamemode Conditionals#############################


def _is_gamemode(modes, context=None):
    """
    Is the robot in one of the given gamemodes?
    """
    if not (isinstance(modes, tuple) or isinstance(modes, list)):
        modes = [modes, ]

    if context is None:
        context = get_context()
    data, lock = context.get_interface_data(context_data_key)

    return data.get("mode", "") in modes


def is_enabled():
    """
    :returns: True if the robot is Enabled.
    """
    return _is_gamemode((AUTONOMOUS, TELEOPERATED))


def is_teleop():
    """
    :returns: True if the robot is in Teleoperated mode.
    """
    return _is_gamemode(TELEOPERATED)


def is_autonomous():
    """
    :returns: True if the robot is in Autonomous mode.
    """
    return _is_gamemode(AUTONOMOUS)


def is_disabled():
    """
    :returns: True if the robot is Disabled.
    """
    return _is_gamemode(DISABLED)

#Gamemode Wait Coroutines############################


@asyncio.coroutine
def wait_for_gamemode(modes, context=None):
    """
    Waits until one of the indicated modes is set.

    This is an asyncio coroutine.

    :param modes: One or more modes to wait for.
    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """

    if not (isinstance(modes, tuple) or isinstance(modes, list)):
        modes = [modes, ]

    if context is None:
        context = get_context()
    data, lock = context.get_interface_data(context_data_key)

    event = asyncio.Event()

    for mode in modes:
        if mode == data.get("mode", ""):
            return
        events_key = "events-mode-" + str(mode)
        if events_key not in data:
            data[events_key] = list()

        data[events_key].append(event)

    yield from event.wait()

    for mode in modes:
        events_key = "events-mode-" + str(mode)
        data[events_key].remove(event)


@asyncio.coroutine
def wait_for_disabled(context=None):
    """
    Waits until the disabled mode is set.

    This is an asyncio coroutine.

    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """
    yield from wait_for_gamemode(DISABLED, context=context)


@asyncio.coroutine
def wait_for_teleop(context=None):
    """
    Waits until the teleoperated mode is set.

    This is an asyncio coroutine.

    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """
    yield from wait_for_gamemode(TELEOPERATED, context=context)


@asyncio.coroutine
def wait_for_autonomous(context=None):
    """
    Waits until the autonomous mode is set.

    This is an asyncio coroutine.

    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """
    yield from wait_for_gamemode(AUTONOMOUS, context=context)


@asyncio.coroutine
def wait_for_enabled(context=None):
    """
    Waits until either the autonomous mode or the teleoperated mode is set.

    This is an asyncio coroutine.

    :param context: The optional context to use for data storage. If none, the current
        thread’s context will be used.
    """
    yield from wait_for_gamemode((TELEOPERATED, AUTONOMOUS), context=context)

#Gamemode Coroutine Decorators##########################


def _gamemode_task(f, gamemodes):
    @asyncio.coroutine
    def wrapper_func(*args, **kwargs):
        while True:
            yield from wait_for_gamemode(gamemodes)
            yield from f(*args, **kwargs)
            #Give the rest of the robot a moment to breath before looping
            yield from asyncio.sleep(.05)
    wrapper_func.autorun = True
    return wrapper_func


def disabled_task(f):
    """
    A decorator function that causes the decorated coroutine to run continually
    when the robot is Disabled.
    """
    return _gamemode_task(f, DISABLED)


def teleop_task(f):
    """
    A decorator function that causes the decorated coroutine to run continually
    when the robot is in Teleoperated Mode.
    """
    return _gamemode_task(f, TELEOPERATED)


def autonomous_task(f):
    """
    A decorator function that causes the decorated coroutine to run continually
    when the robot is in Autonomous Mode.
    """
    return _gamemode_task(f, AUTONOMOUS)


def enabled_task(f):
    """
    A decorator function that causes the decorated coroutine to run continually
    when the robot is Enabled.
    """
    return _gamemode_task(f, (AUTONOMOUS, TELEOPERATED))

#Misc Utils

def gamemode_seconds_elapsed(context=None):
    """
    :param context: The optional context to use for data retrieval.

    :returns: The seconds elapsed since the start of the current gamemode.
    """
    if context is None:
        context = get_context()
    data, lock = context.get_interface_data(context_data_key)
    current_time = wpilib.Timer.getFPGATimestamp()
    return current_time - data.get("last_change", current_time)




"""
interfaces.events is the first default interface mechanism. It provides a
system for referencing and using asyincio events by name.
"""
from functools import partial
from asyncio import Event

from ..context import get_context

context_datastore_key = "events"


def get_event(id, context=None):
    """
    Retrieves an asyncio event object by name, creating one if necessary.

    This method is thread-safe.

    :param id: The string ID of the event to get, or to create.
    :param context: The optional context to use. Otherwise the current
        thread’s context will be used.

    :returns: The asyncio event corresponding to the given id.
    """
    if context is None:
        context = get_context()
    evdata, lock = context.get_interface_data(context_datastore_key)
    with lock:
        if id not in evdata:
            evdata[id] = Event(loop=context.get_event_loop())
        return evdata[id]


def set_event_threadsafe(id, context=None):
    """
    Schedules :meth:`set_event()` to be run in the current context's
    event loop.

    This method is thread-safe.

    :param id: The string ID of the event to set.
    :param context: The optional context to use. Otherwise the current
        thread’s context will be used.
    """
    if context is None:
        context = get_context()
    context.get_event_loop().call_soon_threadsafe(partial(set_event, id, context))


def set_event(eid, context=None):
    """
    Triggers :meth:`set()` on the asyncio event indicated by id, creating
    one if necessary.

    :param id: The string ID of the event to set.
    :param context: The optional context to use. Otherwise the current
        thread’s context will be used.
    """
    if context is None:
        context = get_context()
    evdata, lock = context.get_interface_data(context_datastore_key)
    with lock:
        if eid not in evdata:
            evdata[eid] = Event()
        evdata[eid].set()


def clear_event_threadsafe(eid, context=None):
    """
    Schedules :meth:`clear_event()` to be run in the current context's
    event loop.

    This method is thread-safe.

    :param id: The string ID of the event to clear.
    :param context: The optional context to use. Otherwise the current
        thread’s context will be used.
    """
    if context is None:
        context = get_context()
    context.get_event_loop().call_soon_threadsafe(partial(clear_event, eid, context))


def clear_event(eid, context=None):
    """
    Triggers :meth:`clear()` on the asyncio event indicated by id, creating
    one if necessary.

    :param id: The string ID of the event to clear.
    :param context: The optional context to use. Otherwise the current
        thread’s context will be used.
    """
    if context is None:
        context = get_context()
    evdata, lock = context.get_interface_data(context_datastore_key)
    with lock:
        if eid not in evdata:
            evdata[eid] = Event()
        evdata[eid].clear()



import wpilib
from robotpy_ext.misc import asyncio_policy
from os.path import join
from .. import Context, ConfigManager
from ..interfaces import gamemode


class YetiRobot(wpilib.IterativeRobot):
    """
    This is a standard robot class that uses the full stack of yeti.
    """

    config_dir = ""
    config_file = "mods.conf"

    #Patch the asyncio policy to use wpilib.Timer.getFPGATime() rather than time.monotonic()
    asyncio_policy.patch_asyncio_policy()

    def robotInit(self):
        #First get the context initialized, as ModuleLoaders use it to run
        self.context = Context()
        self.context.start()

        #Then use a ConfigManager to load modules specified in a configuration file
        config_manager = ConfigManager()
        config_manager.parse_config(join(self.config_dir, self.config_file))
        config_manager.load_startup_mods(self.context)

    def teleopInit(self):
        gamemode.set_gamemode(gamemode.TELEOPERATED, context=self.context)

    def disabledInit(self):
        gamemode.set_gamemode(gamemode.DISABLED, context=self.context)

    def autonomousInit(self):
        gamemode.set_gamemode(gamemode.AUTONOMOUS, context=self.context)




import yeti
import time
import asyncio
import os
import json
import traceback
import logging
from aiohttp import web

try:
    from yeti.version import __version__
except:
    __version__ = "master"

class WebUILoggerHandler(logging.Handler):

    def __init__(self, handler_func):
        self.handler_func = handler_func
        super().__init__()

    def emit(self, record):
        self.handler_func(record)

class WebUI(yeti.Module):
    """
    A pre-loaded module that provides an elegant interface with which to manage loaded modules.
    """

    def module_init(self):
        self.context = yeti.get_context()
        self.file_root = os.path.join(os.path.dirname(os.path.realpath(__file__)), "resources")
        self.start_coroutine(self.init_server())
        self.yeti_logger = logging.getLogger("yeti")
        def er_hdl(msg):
            self.context.thread_coroutine(self.error_handler(msg))
        logger_handler = WebUILoggerHandler(er_hdl)
        self.yeti_logger.addHandler(logger_handler)

        self.messages = list()

    next_error_id = 0


    @asyncio.coroutine
    def error_handler(self, message):
        yield from self.clean_messages()
        message_data = {"time": time.monotonic(), "message": message.getMessage(), "level": message.levelname, "id": self.next_error_id}
        self.next_error_id += 1
        self.messages.append(message_data)

    @asyncio.coroutine
    def clean_messages(self):
        #Clean all timed-out messages
        current_time = time.monotonic()
        for message in self.messages[:]:
            if message["time"] + 10 < current_time:
                self.messages.remove(message)

    @asyncio.coroutine
    def json_handler(self, request):
        data_structure = dict()
        data_structure["version"] = __version__
        yield from self.clean_messages()
        data_structure["messages"] = self.messages
        data_structure["next_mid"] = self.next_error_id
        data_structure["modules"] = list()
        for modname in self.context.loaded_modules:
            mod_object = self.context.loaded_modules[modname]
            mod_data = dict()
            mod_data["subsystem"] = modname
            mod_data["description"] = mod_object.__doc__
            if hasattr(mod_object, "loader"):
                mod_data["filename"] = mod_object.loader.module_path
                mod_data["status"] = "Loaded"
                mod_data["fallbacks"] = mod_object.loader.fallback_list
            data_structure["modules"].append(mod_data)
        text = json.dumps(data_structure, allow_nan=False)
        return web.Response(body=text.encode("utf-8"))

    @asyncio.coroutine
    def command_handler(self, request):
        commands = {"load": self.load_command, "load_config": self.load_config, "unload": self.unload_command, "reload": self.reload_command}
        data = yield from request.post()
        try:
            text = yield from commands[data["command"]](data["target"])
        except Exception as e:
            self.logger.error(str(e) + "\n" + traceback.format_exc())
            text = str(e)

        return web.Response(body=text.encode("utf-8"))

    @asyncio.coroutine
    def load_command(self, target):
        if hasattr(self.context, "config_manager"):
            self.context.config_manager.load_module(target, self.context)
        else:
            loader = yeti.ModuleLoader()
            loader.set_context(self.context)
            yield from loader.load_coroutine(target)
        return "Successfully loaded " + target

    @asyncio.coroutine
    def unload_command(self, target):
        yield from self.context.unload_module_coroutine(target)
        return "Successfully unloaded " + target

    @asyncio.coroutine
    def reload_command(self, target):
        all_mods = self.context.get_modules()
        if target == "all":
            target_mods = [all_mods[mod] for mod in all_mods]
        else:
            target_mods = [all_mods[target], ]
        for mod in target_mods:
            if hasattr(mod, "loader"):
                yield from mod.loader.reload_coroutine()
        return "Successfully reloaded " + target

    @asyncio.coroutine
    def load_config(self, path):
        for module in self.context.get_modules():
            yield from self.context.unload_module_coroutine(module)
        if not hasattr(self.context, "config_manager"):
            self.context.config_manager = yeti.ConfigManager()
        self.context.config_manager.parse_config(path)
        self.context.config_manager.load_startup_mods(self.context)
        return "Successfully reloaded config file " + path

    @asyncio.coroutine
    def forward_request(self, request):
        return web.HTTPFound("/index.html")

    @asyncio.coroutine
    def init_server(self):
        app = web.Application()
        app.router.add_route("GET", "/api/json", self.json_handler)
        app.router.add_route("POST", "/api/command", self.command_handler)
        app.router.add_route("GET", "/", self.forward_request)
        app.router.add_static("/", self.file_root)
        self.srv = yield from self.event_loop.create_server(app.make_handler(), port=8080)

        self.logger.info("Yeti WebUI started at  http://127.0.0.1:8080/index.html")

    def module_deinit(self):
        self.srv.close()


import threading
import asyncio
import logging
import traceback

from .hook_server import HookServer


_contexts = dict()

def set_context(context):
    """
    Sets the context for the current thread.

    :param context: An instance of Context to be saved
    """
    _contexts[threading.current_thread()] = context

def get_context():
    """
    :returns: the context set for that thread.
    """
    return _contexts[threading.current_thread()]

class Context(HookServer):
    """
    This hosts an asyncio event loop in a thread, and contains mechanisms for loading and unloading modules.
    """
    _event_loop = None
    _thread = None

    def __init__(self):
        super().__init__()
        self.loaded_modules = dict()
        self.interface_data = dict()
        self.interface_data_lock = threading.RLock()
        self._event_loop = asyncio.new_event_loop()
        self.logger = logging.getLogger(name="yeti." + self.__class__.__name__)

    def thread_coroutine(self, coroutine, logger=None):
        """
        Schedules coroutine to be run in the context's event loop.
        This function is thread-safe.

        :param coroutine: The coroutine to schedule
        """
        if logger is None:
            logger = self.logger
        self._event_loop.call_soon_threadsafe(asyncio.async, self._error_net(coroutine, logger))

    @asyncio.coroutine
    def _error_net(self, coro, log):
        try:
            yield from coro
        except Exception as e:
            self.logger.error(str(e) + "\n" + traceback.format_exc())

    def get_event_loop(self):
        """
        :returns: The context's event loop
        """
        return self._event_loop

    def get_interface_data(self, interface_id):
        """
        Returns a data tuple stored for the identifier interface_id, creating one if necessary.

        :param interface_id: The string identifier for the data tuple to return

        :returns: A data tuple of type (:class:`dict()` , :class:`threading.RLock()`)
        """
        with self.interface_data_lock:
            if interface_id not in self.interface_data:
                self.interface_data[interface_id] = (dict(), threading.RLock())
            data = self.interface_data[interface_id]
        return data

    def start(self):
        """Spawns a new thread and runs :meth:`.run_forever` in it."""
        if self._thread is None:
            self._thread = threading.Thread(target=self.run_forever)
            self._thread.start()

    def run_forever(self):
        """
        Sets the context for the current thread, and runs the context's event loop forever.
        """
        set_context(self)
        asyncio.set_event_loop(self._event_loop)
        self.call_hook("context_start", self)
        self._event_loop.run_forever()

    def run_for(self, time):
        """
        Sets the context for the current thread, and runs the context's event loop for the specified duration.

        :param time: The time in seconds to run the event loop for.
        """
        set_context(self)
        asyncio.set_event_loop(self._event_loop)
        self.call_hook("context_start", self)
        self._event_loop.run_until_complete(asyncio.sleep(time))

    def stop(self):
        """
        Schedules :meth:`.stop_coroutine` to be run in the context's event loop.
        This method is thread-safe.
        """
        self.thread_coroutine(self.stop_coroutine)

    @asyncio.coroutine
    def stop_coroutine(self):
        """
        Unloads all modules and stops the event loop.
        This method is a coroutine.
        """
        for modname in self.loaded_modules:
            yield from self.unload_module_coroutine(modname)
        self.call_hook("context_stop", self)
        self._event_loop.stop()

    def load_module(self, module):
        """
        Schedules :meth:`.load_module_coroutine` to be run in the context's event loop.
        This method is thread-safe.

        :param module: The module object to load into the context.
        """
        self.thread_coroutine(self.load_module_coroutine(module))

    @asyncio.coroutine
    def load_module_coroutine(self, module):
        """
        Loads module into the context, and errors if we already have one with that name. Triggers :meth:`.start()`.

        :param module: The module object to load into the context.
        """
        if module.name in self.loaded_modules:
            raise ValueError("Already have a module with name {} in this context, cannot add another.")
        self.loaded_modules[module.name] = module
        module.start(loop=self._event_loop)
        self.call_hook("module_load", module)

    def unload_module(self, module_name):
        """
        Schedules :meth:`.unload_module_coroutine()` to be run in the context's event loop.
        This method is thread-safe.

        :param module_name: The the name of the module to be unloaded from the context.
        """
        self.thread_coroutine(self.unload_module_coroutine(module_name))

    @asyncio.coroutine
    def unload_module_coroutine(self, module_name):
        """
        Unloads module_name from the context. Triggers :meth:`.stop()`.

        :param module_name: The name of the module to be unloaded the context.
        """
        if module_name not in self.loaded_modules:
            raise ValueError("Module {} not loaded.".format(module_name))
        self.loaded_modules[module_name].stop()
        self.call_hook("module_unload", self.loaded_modules[module_name])
        del(self.loaded_modules[module_name])

    def get_modules(self):
        return self.loaded_modules.copy()


import asyncio
import logging
import functools
import inspect

from .hook_server import HookServer


class Module(HookServer):
    """
    Provide an interface for bundling and externally controlling asyncio coroutines.
    This is intended to be subclassed to create a module for a specific purpose.
    """
    name = "module"
    event_loop = None

    def __init__(self):
        super().__init__()
        self.name = self.__class__.__name__
        self.logger = logging.getLogger('yeti.' + self.name)
        self.tasks = list()
        self.add_hook("end_task", self._finish_task)
        self.add_hook("init", self.module_init)
        self.add_hook("init", self.autorun_coroutines)
        self.add_hook("deinit", self.module_deinit)

    def module_init(self):
        """A default `module_init` hook used for initializing the module, and starting any coroutines."""
        self.logger.warn("module_init: Override me!")

    def module_deinit(self):
        """A default `module_deinit` hook used for freeing any used resources."""
        pass

    def start(self, loop=None):
        """
        Start module operation. It configures the asyncio event loop to use for runtime, and
        calls the `module_init` hook to get things running.

        :param loop: An optional event loop to use for module run.
        """
        if loop is None:
            self.event_loop = asyncio.get_event_loop()
        else:
            self.event_loop = loop
        try:
            self.call_hook("init", supress_exceptions=False)
        except Exception as e:
            self.call_hook("exception", e)

        self.logger.info("Loaded Module {}.".format(self.name))

    def stop(self):
        """
        This is used to stop module operation. It cancels all running coroutines and calls the `module_deinit`
        hook to stop everything
        """
        self.call_hook("deinit")
        for task in self.tasks:
            task.cancel()
        self.event_loop = None
        self.logger.info("Unloaded Module {}".format(self.name))

    def start_coroutine(self, coroutine):
        """
        Schedule an asyncio coroutine in the module's event loop, and add hooks to handle coroutine failure.

        :param coroutine: The asyncio coroutine to schedule.
        """
        if self.event_loop is None:
            raise ValueError("No event loop setup")
        task = asyncio.async(coroutine)
        task.add_done_callback(functools.partial(self.call_hook, "end_task"))
        self.add_hook("deinit", task.cancel)
        self.call_hook("add_task", task)
        self.tasks.append(task)

    def autorun_coroutines(self):
        for name, obj in inspect.getmembers(self):
            if hasattr(obj, "autorun") and obj.autorun:
                self.start_coroutine(obj())

    def _finish_task(self, fut):
        try:
            if fut.exception():
                if not self.call_hook("exception", fut.exception()):
                    return fut.result()
        except asyncio.CancelledError:
            pass

def autorun_coroutine(func):
    """
    A decorator that sets the coroutine to schedule automatically upon module init.
    """
    func.autorun = True
    return func


import logging
import importlib
import traceback
import inspect
import asyncio
import sys

logger = logging.getLogger('yeti.module_loader')

class ModuleLoadError(Exception):
    """This error is raised for errors during module load"""
    def __init__(self, name, message):
        super().__init__(name + ": " + message)


class ModuleUnloadError(Exception):
    """This error is raised for errors during module unload"""
    def __init__(self, name, message):
        super().__init__(name + ": " + message)


class ModuleLoader(object):
    """
    This dynamically imports modules from the filesystem and loads them into a context.
    This also contains mechanisms for loading fallback modules upon module failure.
    """
    def __init__(self):

        self.fallback_list = list()
        self.fallback_index = 0

        self.module_path = ""
        self.module_name = ""
        self.module_object = None
        self.module_import = None
        self.module_context = None

        self.logger = logger

    def set_context(self, context):
        """
        Sets the context for the module loader to load modules into.

        :param context: The context to use for the module
        """
        self.module_context = context


    def get_context(self):
        """
        :returns: The context currently set for the loader.
        """
        if self.module_context is None:
            raise ValueError("No context set.")
        return self.module_context

    def get_module(self):
        """
        :returns: The currently loaded module object.
        """
        return self.module_object

    def reload(self):
        """
        Schedules reload_coroutine to be run in the context currently set for the module_loader.
        This method is thread-safe.
        """
        self.get_context().thread_coroutine(self.reload_coroutine())

    @asyncio.coroutine
    def reload_coroutine(self):
        """
        This unloads any existing module object and loads the next available module in the callback list.
        """
        yield from self.load_coroutine()

    def add_fallback(self, fallback):
        """
        Add a filename to the fallback list.

        :param fallback: The filename to add to the fallback list.
        """
        self.fallback_list.append(fallback)

    def set_fallback(self, fallback_list):
        """
        Replaces the current fallback list with a copy of fallback_list.

        :param fallback_list: The list of filenames to use for the fallback list.
        """
        self.fallback_list = fallback_list[:]

    @asyncio.coroutine
    def load_coroutine(self, module_path=None):
        """
        Loads a module into the loader, taking the filename from the top of the fallback_list.

        Prior to module loading, it will unload any loaded module.

        If given module_path, it will try to find it in the current fallback list. If it fails, it will create a new
        fallback_list with module_path.

        It then iterates through fallback_list, loading modules until one completes successfully.

        :param module_path: The (optional) module filename to use.
        """

        #Start by unloading any previously loaded module
        yield from self.unload_coroutine()

        #Setup module fallback lists

        #If a module was actually specified, try to find it in fallback_list, otherwise,
        # wipe fallback_list and add module_path
        if module_path is not None:
            if module_path in self.fallback_list:
                self.fallback_index = self.fallback_list.index(module_path)
            else:
                self.fallback_list = [module_path]
                self.fallback_index = 0

        #Loop through fallback_list and try to find a module that will load.
        while True:

            #Do we have any more to try?
            if self.fallback_index >= len(self.fallback_list):
                raise ModuleLoadError(self.module_name, "No files left to try and load in the fallback list")

            #Lets try this module, the one selected by fallback_index
            file_to_load = self.fallback_list[self.fallback_index]
            try:

                logger.debug("Loading " + file_to_load)

                #Reload the module if it is already loaded, otherwise import it anew..
                if file_to_load in sys.modules:
                    self.module_import = sys.modules[file_to_load]
                    importlib.reload(self.module_import)
                else:
                    self.module_import = importlib.import_module(file_to_load)

                #Get the module class
                module_class = None
                for name, obj in inspect.getmembers(self.module_import):
                    if inspect.isclass(obj) and hasattr(obj, "module_init"):
                        module_class = obj
                        break
                else:
                    raise ModuleLoadError("Loader", "No compatible module class found in " + file_to_load)

                #Initialize the actual module object
                self.module_object = module_class()

                #Get the module's name and file name
                self.module_name = self.module_object.name
                self.module_path = file_to_load
                self.logger = self.module_object.logger

                #Add control hooks
                self.module_object.add_hook("exception", self._exception_handler)

                #embed self
                self.module_object.loader = self

                #Add module to the current context:
                yield from self.module_context.load_module_coroutine(self.module_object)

                #Yay, we must have been successful!
                break

            except Exception as e:
                #Oops, something happened. We must try the next one on the fallback list!
                self.logger.error("Error loading module: " + file_to_load + ": " + str(e) + "\n" + traceback.format_exc())
                self.fallback_index += 1

    def load(self, module_path=None):
        """Schedules :meth:`load_coroutine` to be run."""
        self.get_context().thread_coroutine(self.load_coroutine(module_path))

    @asyncio.coroutine
    def unload_coroutine(self):
        """Unload the currently loaded module"""
        if self.module_object is not None:
            yield from self.module_context.unload_module_coroutine(self.module_name)
            self.module_object = None

    def unload(self):
        """Schedules :meth:`unload_coroutine` to be run."""
        self.get_context().thread_coroutine(self.unload_coroutine())

    def _exception_handler(self, exception):
        #Oops, something happened
        self.logger.error("Error in module run: {}: {}\n {}".format(self.module_path, str(exception), "".join(traceback.format_tb(exception.__traceback__))))

        #Try to load a replacement module.
        try:
            self.replace_faulty()
        except ModuleLoadError as e:
            logger.error(e)

    @asyncio.coroutine
    def replace_faulty_coroutine(self):
        """Replace a faulty module with the next in line, aka increment fallback_index and trigger load()"""
        self.fallback_index += 1
        try:
            yield from self.load_coroutine()
        except Exception as e:
            self.logger.error("Error replacing faulty module: " + str(e))

    def replace_faulty(self):
        """Schedules :meth:`replace_faulty_coroutine` to be run."""
        self.get_context().thread_coroutine(self.replace_faulty_coroutine())


import logging

from .module_loader import ModuleLoader

logger = logging.getLogger('yeti.ConfigManager')


class ConfigurationError(Exception):
    pass


class ConfigManager(object):
    """
    Uses instances of :class:`ModuleLoader` to load modules from reference lists in a configuration file.
    """

    _STARTUP_MOD_SECTION = "StartupMods"
    _config_path = ""

    def __init__(self):
        self.config_structure = None
        self.module_loaders = dict()

    def load_startup_mods(self, context):
        """
        Find all modules in the "StartupMods" section of the config file, and load them with instances of :class:`ModuleLoader`
        into the specified context.

        :param context: The context to load the modules into.
        """

        if self.config_structure is None:
            raise ConfigurationError("No config file loaded.")
        for module_name in self.config_structure[self._STARTUP_MOD_SECTION]:
            self.load_module(module_name, context)

    def load_module(self, name, context):
        """
        This uses a loaded config file to generate a fallback list and use a :class:`ModuleLoader` to load the module.

        :param name: The name reference of the module to load.
        :param context: The context to load the module into.

        :returns: The created :class:`ModuleLoader`
        """

        #Add reference to context for introspection.
        context.config_manager = self

        if self.config_structure is None:
            fallback_list = [name]
            fallback_index = 0

        #is it a subsystem name?
        elif name in self.config_structure:
            fallback_list = self.config_structure[name]
            fallback_index = 0

        #no? must be a filename then.
        else:
            #Search for filename in loaded config
            for subsystem_config in self.config_structure:
                if subsystem_config != self._STARTUP_MOD_SECTION and name in self.config_structure[subsystem_config]:
                    #We found it! set the fallback list
                    fallback_list = self.config_structure[subsystem_config]
                    fallback_index = fallback_list.index(name)
                    break

            #If we still don't have a fallback list, just make one up and go!
            else:
                fallback_list = [name]
                fallback_index = 0

        module_loader = ModuleLoader()
        module_loader.set_context(context)
        module_loader.fallback_list = fallback_list
        module_loader.fallback_index = fallback_index
        module_loader.load()
        self.module_loaders[name] = module_loader
        return module_loader

    def parse_config(self, path):
        """
        Parse the config file.

        :param path: The file path of the config file to parse.

        :returns: The dictionary of the parsed config file.
        """

        if path == "":
            path = self._config_path
        self._config_path = path

        #Open the file
        f = open(path)
        section = None
        parsed_config = dict()

        #for each line in file:
        for line in f:
            #Get rid of extra spaces and carriage-returns
            line = line.rstrip('\r\n')

            #If there is a comment on the line, get rid of everything after the comment symbol and trim whitespace
            #Example: hi there #This is a comment
            if "#" in line:
                line, comment = line.split("#", 1)
                line = line.strip()

            #If there is a section header on the line, figure out what it's name is, and save it
            if "[" in line:
                #Example: [StartupMods]
                section = line.split("[", 1)[1].split("]", 1)[0]
                parsed_config[section] = list()

            #If there is no section header, than the line must contian data, so save it under the current section
            else:
                if line is not "":
                    parsed_config[section].append(line)

        #Message the system
        logger.info("Finished parsing " + path)
        self.config_structure = parsed_config
        return parsed_config



class Referee:
    """
    This is for keeping track of used wpilib references, and freeing them if their containing modules get unloaded.
    """

    def __init__(self, module):
        self.module = module
        self.refs = list()
        self.hook = module.add_hook("deinit", self.free_all)

    def watch(self, wpilib_object):
        """Keep track of a wpilib object."""
        if hasattr(wpilib_object, "free"):
            self.refs.append(wpilib_object)

    def free_all(self):
        """Frees all watched wpilib references."""
        for wpilib_object in self.refs:
            wpilib_object.free()



#!/usr/bin/env python3

from os.path import dirname, exists, join
import sys, subprocess
from setuptools import setup, find_packages
from distutils.command.install import INSTALL_SCHEMES

for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

setup_dir = dirname(__file__)
base_package = 'yeti'
git_dir = join(setup_dir, ".git")
version_file = join(setup_dir, base_package, 'version.py')

# Automatically generate a version.py based on the git version
if exists(git_dir):
    p = subprocess.Popen(["git", "describe", "--tags", "--long", "--dirty=-dirty"],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    # Make sure the git version has at least one tag
    if err:
        print("Error: You need to create a tag for this repo to use the builder")
        sys.exit(1)

    # Convert git version to PEP440 compliant version
    # - Older versions of pip choke on local identifiers, so we can't include the git commit
    v, commits, local = out.decode('utf-8').rstrip().split('-', 2)
    if commits != '0' or '-dirty' in local:
        v = '%s.post0.dev%s' % (v, commits)

    # Create the version.py file
    with open(join(setup_dir, base_package, 'version.py'), 'w') as fp:
        fp.write("# Autogenerated by setup.py\n__version__ = '{0}'".format(v))

if exists(version_file):
    with open(version_file, 'r') as fp:
        exec(fp.read(), globals())
else:
    __version__ = "master"

with open(join(setup_dir, 'README.rst'), 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name='yeti',
    version=__version__,
    description="Yeti",
    long_description=long_description,
    author='Christian Balcom',
    author_email='robot.inventor@gmail.com',
    url='https://github.com/Team4819/yeti',
    keywords='frc first robotics asyncio',
    packages=find_packages(),
    install_requires=["aiohttp", ],
    include_package_data=True,
    license="BSD License",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Topic :: Scientific/Engineering"]
    )




import logging


class Hook(object):
    """
    This stores a reference to a method, and can be used to remove the hook's reference from the source :class:`HookServer`.
    Hook objects are generally only created by a :class:`HookServer`.
    """

    def __init__(self, hook_server, hook_name, func):
        """
        :param hook_server: The instance of :class:`.HookServer` that uses this hook.
        :param hook_name: The name of the method to create a hook for.
        :param func: The function to use with the hook.
        """
        self.func = func
        self.hook_server = hook_server
        self.hook_name = hook_name

    def call(self, *args, **kwargs):
        """
        Calls the hook function with the given parameters
        """
        self.func(*args, **kwargs)

    def unset(self):
        """
        Removes the trigger from hook_server
        """
        self.hook_server.remove_hook(self)


class HookServer(object):
    """
    This provides a convenient hook registration mechanism.
    """

    def __init__(self):
        self.logger = logging.getLogger("yeti." + self.__class__.__name__)
        self._hooks = dict()

    def add_hook(self, hook_name, callback):
        """
        Adds a hooks to be run when method hook_name is called.

        :param hook_name: The name of the method to create a hook for.
        :param callback: The method to use for the hook.

        :returns: The new instance of :class:`.Hook`
        """
        hook = Hook(self, hook_name, callback)
        if hook_name not in self._hooks:
            self._hooks[hook_name] = list()
        self._hooks[hook_name].append(hook)
        return hook

    def call_hook(self, hook_name, *args, supress_exceptions=True, **kwargs):
        """
        Calls all hooks that have been added to this :class:`.HookServer` for the method hook_name with all extra
        provided parameters.

        :param hook_name: The name of the method with hooks to be called.
        :param supress_exceptions: Whether or not to suppress raised exceptions. Defaults to true.

        :returns: True if at least one hook was successfully called.
        """
        retval = False
        if hook_name in self._hooks:
            for hook in self._hooks[hook_name]:
                try:
                    hook.call(*args, **kwargs)
                    retval = True
                except Exception as e:
                    if supress_exceptions:
                        self.logger.exception("Exception on hook '{}' call: {}".format(hook_name, e))
                    else:
                        raise e
        return retval

    def remove_hook(self, hook):
        """
        Removes hook from the instance of :class:`.HookServer`

        :param hook: The :class:`.Hook` object to remove.

        :returns: Weather or not removal was successful.
        """
        if hook in self._hooks[hook.hook_name]:
            self._hooks[hook.hook_name].remove(hook)
            return True
        return False




import asyncio

class Button:

    def __init__(self, joystick, button):
        self.joystick = joystick
        self.button = button

    @asyncio.coroutine
    def until_rising(self):
        last_val = True
        while True:
            val = self.joystick.getRawButton(self.button)
            if val and not last_val:
                return
            last_val = val
            yield from asyncio.sleep(.02)

    @asyncio.coroutine
    def until_falling(self):
        last_val = False
        while True:
            val = self.joystick.getRawButton(self.button)
            if not val and last_val:
                return
            last_val = val
            yield from asyncio.sleep(.02)

    @asyncio.coroutine
    def until_pressed(self):
        while not self.joystick.getRawButton(self.button):
            yield from asyncio.sleep(.02)